from .commands import CommandResult, create_commander, group, keyword, optional, var  # noqa: F401
from .validation import validate  # noqa: F401
from .transformation import transform  # noqa: F401
from .error import CommandError, ItemNotFoundError, ParseError  # noqa: F401



from .. import create_commander
from ..events import ConsoleExitEvent

command = create_commander('quit')


@command('exit')
def exit():
    """ exit the interpreter """
    raise ConsoleExitEvent()


@command('quit')
def quit():
    """ exit the interpreter """
    raise ConsoleExitEvent()



import fcntl
import os
import signal
import sys
import termios
import urllib.parse


from ..utils import FragileStreamHandler


def _parse_value_from_string(value, target_type):
    if target_type is bool:
        if value.lower() in {"0", "off", "no", "false"}:
            return False
        elif value.lower() in {"1", "on", "yes", "true"}:
            return True
        else:
            raise ValueError
    elif target_type is int:
        # may raise ValueError
        return int(value)
    else:
        raise NotImplementedError


def get_interface_handler(specification, logger):
    if specification is None:
        return LocalInterface()
    parsed = urllib.parse.urlparse(specification)
    kwargs = dict(urllib.parse.parse_qsl(parsed.query))
    if parsed.scheme == "tty":
        parser_map = {"reconnect_on_hangup": bool, "baudrate": int}
        handler = TTYInterface
        args = (parsed.path, )
    elif parsed.scheme == "telnet":
        parser_map = {"port": int}
        if ":" in parsed.path:
            host, kwargs["port"] = parsed.path.split(":", 1)
        else:
            host = parsed.path
        handler = TelnetInterface
        args = (host, )
    # parse and convert values
    for key, target_type in parser_map.items():
        if key in kwargs:
            try:
                kwargs[key] = _parse_value_from_string(kwargs[key], target_type)
            except ValueError as exc:
                raise InvalidInterfaceSpecificationError(
                    "Failed to identify {} value of '{}': {}"
                    .format(target_type.__name__, key, kwargs[key])) from exc
    # create an instance of the handler
    try:
        return handler(logger, *args, **kwargs)
    except TypeError as exc:
        raise InvalidInterfaceSpecificationError("Failed to instantiate terminal handler ({}): {}"
                                                 .format(handler.__class__.__name__, exc)) from exc


class InvalidInterfaceSpecificationError(Exception):
    """ Indicate an invalid terminal specification (e.g. unknown scheme or missing arguments) """


class BaseInterface:
    """ interface implementations are supposed to reconfigured sys.stdin and sys.stdout

    The details of the interfaces should be hidden well, in order to allow its users to keep using
    the builtins "input()" and "print()".
    """

    def __init__(self, logger):
        self.logger = logger

    def connect(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError


class LocalInterface(BaseInterface):
    """ communicate directly with stdin and stdout """

    def connect(self):
        pass

    def close(self):
        pass

    def __str__(self):
        return ("Local")


class TTYInterface(BaseInterface):

    def __init__(self, logger, path, baudrate=115200, reconnect_on_hangup=True):
        super().__init__(logger)
        self.path = path
        self.baudrate = baudrate
        self.is_initialized = False
        self.reconnect_on_hangup = reconnect_on_hangup
        if self.baudrate is not None:
            # Try to import the module early: errors in "connect" are much harder to debug, since
            # they are running in a separate process without stderr.
            import serial
            try:
                self.baudrate_tcattr_value = serial.Serial.BAUDRATE_CONSTANTS[self.baudrate]
            except KeyError:
                InvalidInterfaceSpecificationError("Non-standard baudrates are not supported: {}"
                                                   .format(self.baudrate))

    def connect(self):
        def _sighup_handler(sig, frame):
            self.logger.info("Received SIGHUP: reconnecting to %s", self)
            self.connect()
        if not self.is_initialized:
            # execute all preparations (e.g. forking) for acquiring our terminal later on
            self._become_session_leader()
            if self.reconnect_on_hangup:
                # survive a SIGHUP signal, if requested (useful for USB interfaces)
                self.logger.info("Registring SIGHUP handler for reconnect")
                signal.signal(signal.SIGHUP, _sighup_handler)
            self.is_initialized = True
        self._acquire_controlling_terminal()

    def close(self):
        sys.stdin.close()
        sys.stdout.close()

    @staticmethod
    def _become_session_leader():
        """ replicate the effect of the helper program 'setsid ...' """
        try:
            os.setsid()
        except PermissionError:
            # We need to fork in order to be able to create our own session group.
            if os.fork() != 0:
                # the parent process may die
                sys.exit()
            else:
                os.setsid()

    def _acquire_controlling_terminal(self):
        """ fork the process in order to become session leader and replace stdin/stdout/stderr

        The result should be comparable with running the program via "setsid -w agetty ...".
        """
        # TODO: the flags are just copied from agetty's behaviour
        self.logger.debug("Opening target terminal: %s", self.path)
        dev = os.open(self.path, os.O_RDWR | os.O_NOCTTY | os.O_NONBLOCK | os.O_LARGEFILE)
        self.logger.debug("Closing stdin and stdout")
        for handle in (sys.stdin, sys.stdout):
            try:
                handle.close()
            except OSError:
                # For unknown reasons this seems to happen for stdout from time to time.
                # How to reproduce: disconnect the USB peer in a USB gadget setup for a short very
                # period.
                self.logger.info("Failed to close handle: %s", handle)
        self.logger.debug("Disabling controlling terminal")
        fcntl.ioctl(dev, termios.TIOCSCTTY, 0)
        # duplicate the open handle to stdin and stdout
        self.logger.debug("Assigning terminal to stdin and stdout")
        os.dup2(dev, 0)
        os.dup2(dev, 1)
        # replace the text-based stdin/stdout handles
        sys.stdin = open(0, "rt", buffering=1)
        sys.stdout = open(1, "wt", buffering=1)
        # apply some useful settings for an interactive terminal
        self.logger.debug("Applying interactive terminal settings")
        term_settings = termios.tcgetattr(dev)
        # output: new line should also include a carriage return
        term_settings[3] = term_settings[3] | termios.ONLCR
        if self.baudrate is not None:
            term_settings[4] = self.baudrate_tcattr_value
            term_settings[5] = self.baudrate_tcattr_value
        termios.tcsetattr(dev, termios.TCSADRAIN, term_settings)
        os.close(dev)

    def __str__(self):
        return ("TTY (dev={}, baudrate={:d}, reconnect_on_hangup={})"
                .format(self.path, self.baudrate, self.reconnect_on_hangup))


class TelnetInterface(BaseInterface):

    def __init__(self, logger, host, port=None):
        super().__init__(logger)
        raise NotImplementedError


class InterfaceManager:

    def __init__(self, logger):
        self.running_interface_processes = {}
        self.logger = logger

    def run(self, wanted_interfaces, log_filename_by_process=None):
        """ fork processes for each wanted interface

        This function returns for the child processes (i.e.: they may continue using stdin/stdout).
        The parent process (the interface manager) stays alive and never returns.
        """
        if not self.run_and_stop_interfaces(
                wanted_interfaces, log_filename_by_process=log_filename_by_process):
            # we are a managed interface process - simply return
            return
        # we are the manager
        # Ignore all signals that we want to listen for, but store their previous handlers.
        original_signal_handlers = {}
        for one_signal in {signal.SIGUSR1, signal.SIGHUP, signal.SIGTERM, signal.SIGCHLD}:
            # The signal handler signal.SIG_IGN is not appropriate: it would prevent "sigwaitinfo"
            # below from capturing these signals.
            original_signal_handlers[one_signal] = signal.signal(one_signal, lambda *args: None)
        should_finish = False
        while not should_finish:
            try:
                signal_info = signal.sigwaitinfo(set(original_signal_handlers))
            except KeyboardInterrupt:
                self.logger.info("Received CTRL-C")
                should_finish = True
                continue
            if signal_info.si_signo == signal.SIGTERM:
                self.logger.info("Received termination signal")
                should_finish = True
            elif signal_info.si_signo == signal.SIGCHLD:
                # a child process signals its termination
                matched_specs = {spec for spec, pid in self.running_interface_processes.items()
                                 if pid == signal_info.si_pid}
                for spec in matched_specs:
                    self.logger.info("A child signals its termination: %d %s",
                                     signal_info.si_pid, spec)
                    self.running_interface_processes.pop(spec)
                    # in any case: retrieve its result (otherwise it will end up as a zombie)
                    os.waitpid(signal_info.si_pid, os.WNOHANG)
                if not self.running_interface_processes:
                    self.logger.info("All child processes are gone. Going home, too.")
                    should_finish = True
            elif signal_info.si_signo == signal.SIGUSR1:
                # print all currently running processes
                for spec, pid in self.running_interface_processes.items():
                    print("{:d}\t{}".format(pid, spec))
            elif signal_info.si_signo == signal.SIGHUP:
                # Parse the wanted set of interfaces from a given text file and update the list of
                # running processes appropriately.
                raise NotImplementedError(
                    "The 'update interfaces from file' feature is not supported, yet.")
            else:
                raise NotImplementedError("Received unknown signal")
        # kill all child processes
        self.run_and_stop_interfaces(set())
        sys.exit(0)

    def run_and_stop_interfaces(self, wanted_interfaces, log_filename_by_process=None):
        """ kill old processes or start new ones

        @param wanted_interfaces: list of text-based interface specifications
        @param log_filename_by_process: filename template (including "%d" for the PID) to be used
            as a location for storing log message of all child processes. Disabled if undefined.
        """
        running_set = set(self.running_interface_processes)
        obsolete_interfaces = running_set.difference(wanted_interfaces)
        new_wanted_interfaces = wanted_interfaces.difference(running_set)
        if obsolete_interfaces:
            self.logger.info("Terminating processes of obsolete interfaces: %s",
                             obsolete_interfaces)
        # kill obsolete child processes
        for obsolete in obsolete_interfaces:
            child_pid = self.running_interface_processes.pop(obsolete)
            try:
                os.kill(child_pid, signal.SIGTERM)
            except OSError:
                self.logger.warning("Failed to kill obsolete child process: %d (%s)",
                                    child_pid, obsolete)
        # parse the specifications first (reduced risk of leaving a broken mess)
        if new_wanted_interfaces:
            self.logger.info("Starting processes for new interfaces: %s", new_wanted_interfaces)
        interface_handlers = {}
        for new_spec in new_wanted_interfaces:
            try:
                handler = get_interface_handler(new_spec, self.logger)
            except InvalidInterfaceSpecificationError as exc:
                self.logger.error("Failed to parse interface specification ('%s'): %s",
                                  new_spec, exc)
                sys.exit(1)
            else:
                interface_handlers[new_spec] = handler
        # start new processes and memorize their PIDs
        for spec, handler in interface_handlers.items():
            child_pid = os.fork()
            if child_pid == 0:
                # we are the child
                # Release the (shared) error output file.
                sys.stderr.close()
                try:
                    os.close(2)
                except OSError:
                    pass
                sys.stderr = open(os.path.devnull, "wt")
                # connect to our target interface
                handler.connect()
                # optionally redirect stderr for logging
                if log_filename_by_process:
                    sys.stderr.close()
                    sys.stderr = open(log_filename_by_process % os.getpid(), "wt")
                # our logging handler also influences the logging handler of the interface
                self.reconfigure_logging()
                self.logger.info("Spawned process %d for interface: %s", os.getpid(), handler)
                # there is nothing more to be prepared by us - the logic handler may take over
                return False
            else:
                # we are the parent
                self.running_interface_processes[spec] = child_pid
        # only the parent ends up here
        return True

    def reconfigure_logging(self):
        """ remove all existing log handlers and add a stream handler to stderr (if open) """
        self.logger.debug("Reconfiguring logging")
        for handler in self.logger.handlers:
            self.logger.removeHandler(handler)
        if sys.stderr.closed:
            # there is no output target available - do nothing
            pass
        else:
            self.logger.addHandler(FragileStreamHandler(sys.stderr))
        self.logger.debug("Finished reconfiguring logging")



import os
import re
from collections import OrderedDict
from itertools import groupby
from textwrap import wrap, indent

from ..commands import create_commander, var, CommandResult, filter_root_commands, \
    filter_public_commands
from ..transformation import flatten, unique
from ..utils import join_generator_string, TaggedString
from ..utils import strip_tags, read_object_path
from ..error import CommandError

WRAP_WIDTH = 80

command = create_commander('help', description='List and describe all available commands.')


def _indent_no_first():
    first = [True]

    def callback(line):
        if not line.strip():
            return False
        if first[0]:
            first[0] = False
            return False
        else:
            return True
    return callback


def _wrap_indent(text, prefix, text_width=WRAP_WIDTH):
    wrapped_text = os.linesep.join(wrap(_str(text), width=text_width))
    return indent(wrapped_text, prefix)


def _str(text):
    return text.strip() if isinstance(text, str) else ''


def render_description(cmd, long_prefix='  - ',  text_width=WRAP_WIDTH):
    doc = re.sub(r'[ ]{2,}', '', _str(cmd.__doc__))
    _indent = '\t' + ' ' * len(long_prefix)

    if doc:
        lines = _wrap_indent(doc, _indent, text_width).split(os.linesep)
        lines[0] = lines[0].replace(_indent, '\t' + long_prefix)
        doc = os.linesep + os.linesep.join(lines)
    for token in cmd.tokens:
        if not token.name:
            continue
        if not token.description and not token.choices:
            continue
        token_name = strip_tags(str(token)) + ': '
        token_doc = os.linesep + _indent + token_name
        token_indent = _indent + ' ' * len(token_name)

        if token.description:
            description = os.linesep.join(wrap(token.description, text_width - len(token_indent)))
            token_description = indent(description, token_indent, _indent_no_first())
            token_doc += token_description
        if token.choices:
            if token.description:
                token_doc += os.linesep + token_indent
            token_doc += '%s' % ' | '.join(sorted(map(str, token.choices)))
        doc += token_doc

    return TaggedString.help(doc) if doc else ''


@join_generator_string()
def describe_command_list(commands: dict):
    yield TaggedString.label('Supported Commands')
    for label, command in commands.items():
        yield '\t%s' % label
    yield 'Use "%s" for a detailed help on individual commands' % str(help_command)


@join_generator_string()
@command.inject(terminal='terminal')
def describe_command(terminal, all_commands, root):
    yield TaggedString.header('%s command' % str(root))
    if root.__commander__.__doc__:
        yield TaggedString.help(_wrap_indent(root.__commander__.__doc__, ''))
    yield ''
    yield TaggedString.label('Supported Subcommands')
    public_commands = filter_public_commands(all_commands)
    sub_commands = [cmd for cmd in public_commands if is_command_root(root, cmd)]
    first = True
    for group, group_commands in groupby(sub_commands, key=lambda cmd: cmd.group):
        if first:
            first = False
        else:
            yield ''

        if group is not None:
            yield '\t%s' % TaggedString.label(group.name)
            if group.description:
                yield TaggedString.help(indent(os.linesep.join(wrap(group.description)), '\t  '))
        for cmd in group_commands:
            cmd_str = str(cmd)
            term_width = read_object_path(terminal, 'width', default=None)
            text_width = min(term_width, WRAP_WIDTH) if term_width else WRAP_WIDTH
            doc = render_description(cmd, text_width=text_width)
            yield '\t%s%s' % (cmd_str, doc)


def is_command_root(root, command):
    if root is command:
        return True

    if command.parent is None:
        return False
    else:
        return is_command_root(root, command.parent)


@command('help', is_abstract=True)
def help():
    """ Shows available commands and documentation """
    pass


@command(var('command', is_optional=True), parent=help)
@command.inject(commands='__commands[]')
@command.transform(commands=[flatten, unique])
def help_command(commands, command=None):
    root_commands = filter_root_commands(commands)
    command_map = OrderedDict(
        (str(cmd), cmd) for cmd in sorted(root_commands, key=lambda cmd: str(cmd))
    )
    if command is None:
        yield CommandResult(describe_command_list(command_map))
    else:
        try:
            yield CommandResult(describe_command(commands, command_map[command]))
        except KeyError as exc:
            raise CommandError('No help for command "%s" available' % command) from exc


@command.inject(commands='__commands[]')
@command.transform(commands=[flatten, unique])
def find_root_commands(commands):
    root_commands = filter_root_commands(commands)
    command_map = OrderedDict(
        (str(cmd), cmd) for cmd in sorted(root_commands, key=lambda cmd: str(cmd))
    )
    return command_map



import argparse
import logging
import os
import readline
import sys
from typing import Sequence

import blinker

from .console import Console
from .console.interfaces import InterfaceManager
from .events import ConsoleExitEvent, StopProcessingEvent
from .utils import autoload, failsafe, FragileStreamHandler, CommandExecutionContext
from .error import CommandError, ShouldThrottleException

logger = logging.getLogger('eliza-runner')

on_result = blinker.signal('result')
on_dispatch = blinker.signal('dispatch')
on_error = blinker.signal('error')

LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}


def initialize_pid_file(path):
    """ create a PID file and take care, that it is safely removed when the program exits """
    def cleanup_pid_file():
        """ remove the PID file, if it contains the current process ID

        This prevents our child processes (which inherit our exit caller via "atexit") from
        removing the parent's PID file.
        """
        try:
            with open(path, 'r') as pid_file:
                stored_pid = int(pid_file.read())
        except (OSError, ValueError):
            pass
        else:
            if stored_pid == os.getpid():
                try:
                    os.unlink(path)
                except OSError:
                    pass

    with open(path, 'w') as pid_file:
        pid_file.write(str(os.getpid()))


class ConsoleRunner:
    def __init__(self) -> None:
        self.cli_args = None
        self.console = None
        self.commander = None

    @property
    def prompt(self):
        return '# '

    def get_arg_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser()
        parser.add_argument('--disable-style', dest='disable_style', action='store_true',
                            help='Disable terminal styling (colors).')
        parser.add_argument('--autoload', dest='modules', action='append',
                            help='autoload a module for the interpreter')
        parser.add_argument('--history', default=None,
                            help='Read history from and write it to the specified file')
        parser.add_argument('--interface', dest='interfaces', action='append',
                            help=('Bind to the given interface(s) and use these for input and '
                                  'output'))
        parser.add_argument('--pid-file', default=None, type=str, help='write process pid to file')
        parser.add_argument('--log-level', dest='log_level', choices=tuple(LOG_LEVELS),
                            default='warning', help='select log verbosity')
        parser.add_argument('--log-filename-by-process', dest='log_filename_by_process', type=str,
                            help=('Store log messages of forked child processes in separate files '
                                  '(e.g. "/var/log/kien/process-%%d.log"). '
                                  'Log storage per process is disabled by default.'))
        parser.add_argument('--ignore-eof', dest='ignore_eof', action='store_true',
                            help='Ignore the EOF control character (commonly: CTRL-D)')
        parser.add_argument('--failsafe', action='store_true',
                            help='keep the application running no matter what '
                                 'runtime exceptions are thrown')
        parser.add_argument('--failsafe-errors', action='store_true',
                            help='in case failsafe is enabled this option will log any unhandled '
                                 'exceptions that are encountered')
        parser.add_argument('--simulate', dest='simulate', action='store_true',
                            help='Show which command would have been executed instead of '
                                 'executing it')
        return parser

    def parse_args(self, args: Sequence[str] = None) -> argparse.Namespace:
        args = sys.argv[1:] if args is None else args
        return self.get_arg_parser().parse_args(args)

    def configure(self) -> None:
        self.cli_args = self.parse_args()
        logger.setLevel(LOG_LEVELS[self.cli_args.log_level])
        logger.addHandler(FragileStreamHandler(sys.stderr))

        # write pid file if requested (we assume that no later forks will happen)
        if self.cli_args.pid_file:
            initialize_pid_file(self.cli_args.pid_file)

        if not self.cli_args.interfaces:
            # Default to the local terminal via stdin/stdout.
            # Implicitly assume, that we do not need the process manager.
            return
        else:
            # Handle all wanted interfaces with a process manager.
            # allow the empty string for "use current stdin/stdout"
            terminal_dev_specifications = set(None if iface == "" else iface
                                              for iface in self.cli_args.interfaces)
            logger.info("Initializing InterfaceManager")
            interface_manager = InterfaceManager(logger)
            # This method call returns once with a fork for each wanted interface.
            # The manager process itself never returns.
            logger.info("Running InterfaceManager")
            interface_manager.run(terminal_dev_specifications,
                                  log_filename_by_process=self.cli_args.log_filename_by_process)

    def run(self) -> None:
        self.configure()

        with Console(lambda: sys.stdout, prompt=self.prompt) as console:
            console.configure_auto(force_disable_style=self.cli_args.disable_style)
            if self.commander is None:
                raise RuntimeError('You must configure a commander before starting run')
            self.commander.provide('console', console)
            self.commander.provide('terminal', console.terminal)
            self.console = console
            if self.cli_args.modules:
                autoload(self.commander, self.cli_args.modules)
            if self.cli_args.history is not None:
                readline.read_history_file(self.cli_args.history)

            def handle_unhandled_exception(exc, *args):
                if self.cli_args.failsafe_errors:
                    logger.error('unhandled exception', exc_info=exc, extra=dict(callargs=args))
                return True

            # start the application loop
            while True:
                @failsafe(enable=self.cli_args.failsafe, callback=handle_unhandled_exception)
                def _process_line():
                    return self._process_line(console, self.commander,
                                              ignore_end_of_file=self.cli_args.ignore_eof)
                try:
                    if not _process_line():
                        if self.cli_args.history is not None:
                            readline.write_history_file(self.cli_args.history)
                        break
                except KeyboardInterrupt:
                    # slow / long operations (or repetitions)
                    # may be interrupted (without quitting)
                    console.linefeed()

    def _process_line(self, console, commander, ignore_end_of_file) -> bool:
        prompt = console.get_prompt()
        try:
            if prompt:
                line = input(prompt)
            else:
                line = input()
        except EOFError:
            console.linefeed()
            if ignore_end_of_file:
                # silently ignore EOF
                return True
            else:
                # request exit
                return False

        if line:
            try:
                # Process results generated by dispatching the provided input line.
                # A new CommandExecutionContext is created for every execution and
                # delegates flow control back to the called generator.
                context = CommandExecutionContext()
                with commander.provide(CommandExecutionContext, context) and context:
                    on_dispatch.send(self, line=line)
                    for result in commander.dispatch(line):
                        # As CommandExecutionContext is a reentrant context manager we can
                        # use it multiple times. In this case we want to give the
                        # dispatched callback control over the dispatch mechanism as well
                        # as console write operations
                        with context:
                            on_result.send(self, result=result)
                            console.send_data(result)
            except ShouldThrottleException as exc:
                logger.error('The output interface could not write data fast enough and the '
                             'command sending the data didn’t implement a back-pressure '
                             'mechanism to decrease the data flow. The command was aborted.',
                             exc_info=exc)
                on_error.send(self, exc=exc)
                console.send_data(exc)
            except CommandError as exc:
                on_error.send(self, exc=exc)
                console.send_data(exc)
            except StopProcessingEvent:
                return True
            except ConsoleExitEvent:
                # the user requested to leave the terminal
                return False
        return True
